#include "seaview.h"
#include "FL/Fl_Group.H"
#include "FL/Fl_Menu.H"
#include "FL/Fl_Menu_.H"
#include "FL/Fl_Browser_.H"
#include "FL/filename.H"
#include "FL/Fl_Button.H"
#include "FL/Fl_Check_Button.H"

/* included functions */
char* fl_file_chooser_save_as(const char* message, const char* pat, 
	const char* fname, void *view);
#ifdef __APPLE__
char* fl_file_chooser_plus(const char* message, const char* pat, 
	const char* fname, int use_only_button);
int MG_GetOutputFName_Popup(char *fname, int maxl, char *dfault, void *data, int defaultformat);
int MG_GetOutputFName_Plus(char *fname, int maxl, char *dfault, char *message, int);
#elif defined(WIN32)
char *MG_win32_file_chooser(const char* message,
        const char* pattern,
        const char* fname,
        const char* defext, // default filename extension or NULL
        bool save, int *pfilterindex);
#endif



// needed functions
extern void set_save_format(SEA_VIEW *view, int val);
#ifdef __APPLE__
extern int set_res_value(const char *name, const char *value);
extern "C" {  
	void MG_FSSpecToPathname (FSSpec *myFSS, char *fname, int maxl);
	void PtoC(const void *in, void *out);
	void CtoP(const void *in, void *out);
#if TARGET_RT_MAC_MACHO
#	define mac_os_x()  1
#else
	int mac_os_x(void);
#endif
	}
#endif

extern int nbr_formats;
extern char *f_format_names[];
extern int printout_block, printout_fontsize;
extern int printout_vary, printout_black;
extern paperformat printout_pageformat;


#ifdef __APPLE__
NavMenuItemSpecArrayHandle formatitems=NULL; /* pour MG_GetOutputFName_Popup et MyNavEventProc_popup */

static pascal void MyNavEventProc_popup (NavEventCallbackMessage callBackSelector,
                                   NavCBRecPtr callBackParms,
                                   NavCallBackUserData callBackUD)
{
NavMenuItemSpec *p;
SEA_VIEW *view;
int num;
OSErr err;

view = (SEA_VIEW *)callBackUD;
if(callBackSelector == kNavCBStart) { /* fixer la valeur montre du menu popup format: */
	HLock((Handle)formatitems);
	p = *formatitems;
	err = NavCustomControl(callBackParms->context, kNavCtlSelectCustomType, 
		p + view->format_for_save ); 
	HUnlock((Handle)formatitems);
	}
else if(callBackSelector == kNavCBPopupMenuSelect) { /* action sur le menu format: */
	p = (NavMenuItemSpec *)(callBackParms->eventData.eventDataParms.param);
	num = p->menuCreator;
	set_save_format(view, num);
	}
return;
}


int MG_GetOutputFName_Popup(char *fname, int maxl, char *dfault, void *data, int defaultformat)
{
    OSErr               anErr = noErr;
    NavReplyRecord      reply;
static NavDialogOptions    dialogOptions;
    OSType              fileTypeToSave = 'TEXT';
    OSType              creatorType;
static NavEventUPP  eventProc;
    NavMenuItemSpec 	*p;
    int  i, rsult = FALSE;

if(formatitems == NULL) { /* initialisation */
    anErr = NavGetDefaultDialogOptions (&dialogOptions);
    if (anErr != noErr) return FALSE;    
    CtoP("seaview: Choose file and format", dialogOptions.windowTitle);
    dialogOptions.dialogOptionFlags &= ~ kNavNoTypePopup;
    dialogOptions.dialogOptionFlags &= ~ kNavAllowStationery;
    formatitems = (NavMenuItemSpecArrayHandle)NewHandle(nbr_formats * sizeof(NavMenuItemSpec));
    if(formatitems == NULL) return FALSE;
    HLock((Handle)formatitems);
    p = *formatitems;
    for(i = 0; i < nbr_formats; i++) {
	p->version = kNavMenuItemSpecVersion;
	CtoP(f_format_names[i], p->menuItemName);
	p->menuCreator = i;  
	p->menuType = 11 ;  /* car les valeurs [-1,10] sont interdites pour menuType */
	p++;
	}
    HUnlock((Handle)formatitems);
    dialogOptions.popupExtension = formatitems;
    eventProc = NewNavEventUPP (MyNavEventProc_popup);
    }
reply.translationNeeded = false;
CtoP(dfault, dialogOptions.savedFileName);
creatorType = kNavGenericSignature;
anErr = NavPutFile(NULL, &reply, &dialogOptions, eventProc, fileTypeToSave, creatorType, data);
if (anErr == noErr && reply.validRecord)
            {
                AEKeyword   theKeyword;
                DescType    actualType;
                Size        actualSize;
                FSSpec      documentFSSpec;

                anErr = AEGetNthPtr(&(reply.selection), 1, typeFSS,
                                    &theKeyword, &actualType,
                                    &documentFSSpec, sizeof(documentFSSpec),
                                    &actualSize );
                if (anErr == noErr)
                {
                    if (reply.replacing)
                    { /* compute the pathname of the pre-existing file */
                        MG_FSSpecToPathname (&documentFSSpec, fname, maxl);
                        rsult = TRUE;
                    }
                    else
                    { /* the chosen file is new, compute pathname to its directory
                    and append new filename */
                        char empty[256];
                        FSSpec pfss;
                        empty[0] = 0;
                   /* make an FSSpec from volume + parentID of documentFSSpec
                   */
                        FSMakeFSSpec(documentFSSpec.vRefNum, documentFSSpec.parID,
                        	(StringPtr) empty, &pfss);
                   /* compute pathname for this FSSpec
                   */
                        MG_FSSpecToPathname (&pfss, fname, maxl);
                        /* append : new filename to this path
                        */
#if TARGET_RT_MAC_MACHO
                        strcat(fname, "/");
#else
                        strcat(fname, ":");
#endif
                        PtoC(documentFSSpec.name, fname + strlen(fname) );
         				anErr = FSpCreate(&documentFSSpec, 'ttxt', 'TEXT', smSystemScript);

                        rsult = TRUE;
                    }
                    anErr = NavCompleteSave(&reply, kNavTranslateInPlace); 
                }
            }
NavDisposeReply(&reply);
if(!rsult) set_save_format((SEA_VIEW *)data, defaultformat);
return rsult;
}


char* fl_file_chooser_save_as(const char* message, const char* pat, 
	const char* fname, void *data)
{
	static char pathname[FL_PATH_MAX];
	SEA_VIEW *view = (SEA_VIEW *)data;
	if( MG_GetOutputFName_Popup(pathname, FL_PATH_MAX, extract_filename((char *)fname), data, 
						view->format_for_save) ) return pathname;
	else return NULL;
}


static pascal void MyNavEventProc_Plus (NavEventCallbackMessage callBackSelector,
                                   NavCBRecPtr callBackParms,
                                   NavCallBackUserData callBackUD)
{
OSErr err;
static Handle gDitlList = NULL;
static DialogItemIndex index;
static DialogRef ref;
static ControlRef var_control = NULL;
Rect rect; Handle h; DialogItemType dtype; Str255 text; ControlRef control;
char text2[256];

if(callBackSelector == kNavCBCustomize) {
	if(callBackParms->customRect.bottom == 0) 
			callBackParms->customRect.bottom = callBackParms->customRect.top + 35;
	if(callBackParms->customRect.right == 0) 
			callBackParms->customRect.right = callBackParms->customRect.left + 425;
	}
else if(callBackSelector == kNavCBStart) {
	gDitlList = GetResource ('DITL', 128);
	if(gDitlList != NULL) err = NavCustomControl(callBackParms->context, 
                            kNavCtlAddControlList, gDitlList);
    err = NavCustomControl(callBackParms->context, kNavCtlGetFirstControlID, &index);
    index++;
    ref = GetDialogFromWindow(callBackParms->window);
	sprintf(text2,"%2d", printout_block);
	CtoP(text2, text);
	GetDialogItem(ref, index + 1, &dtype, &h, &rect);
	SetDialogItemText(h, text);
	sprintf(text2,"%2d", printout_fontsize);
	CtoP(text2, text);
	GetDialogItem(ref, index + 3, &dtype, &h, &rect);
	SetDialogItemText(h, text);
	err = GetDialogItemAsControl(ref, index + 4, &control);
	if(err == 0) SetControl32BitValue(control, printout_black);
	
	err = GetDialogItemAsControl(ref, index + 5, &control);
	if(err == 0) SetControl32BitValue(control, printout_pageformat == 0);
	err = GetDialogItemAsControl(ref, index + 6, &control);
	if(err == 0) SetControl32BitValue(control, printout_pageformat == 1);

	err = GetDialogItemAsControl(ref, index + 7, &var_control);
	if(err == 0) SetControl32BitValue(var_control, printout_vary);
	if(*(int *)callBackUD) ActivateControl(var_control);
	else DeactivateControl(var_control);
	}
else if(callBackSelector == kNavCBTerminate) {
	if(gDitlList != NULL) {
		GetDialogItem(ref, index + 1, &dtype, &h, &rect);
		GetDialogItemText(h, text);
		PtoC(text, text2);
		sscanf(text2,"%d", &printout_block);
		GetDialogItem(ref, index + 3, &dtype, &h, &rect);
		GetDialogItemText(h, text);
		PtoC(text, text2);
		sscanf(text2,"%d", &printout_fontsize);
		set_res_value("printoutfontsize", text2);
		ReleaseResource( gDitlList );
		}
	var_control = NULL;
	}
else if(callBackSelector == kNavCBEvent ) {
	if(var_control != NULL && callBackParms->eventData.itemHit == index + 7) {
		static int count = 0; /* MacOS 9: passe 2 fois a chaque click sur check box */
		if( mac_os_x() || (++count) % 2 == 0 ) {
			printout_vary = (printout_vary + 1) % 2;
			SetControl32BitValue(var_control, printout_vary);
			}
		}
	else if(callBackParms->eventData.itemHit == index + 4) {
		static int count = 0; /* MacOS 9: passe 2 fois a chaque click sur check box */
		if( mac_os_x() || (++count) % 2 == 0 ) {
			printout_black = (printout_black + 1) % 2;
			GetDialogItemAsControl(ref, index + 4, &control);
			SetControl32BitValue(control, printout_black);
			}
		}
	else if(callBackParms->eventData.itemHit == index + 5) {
		if(printout_pageformat == LETTER) {
				GetDialogItemAsControl(ref, index + 5, &control);
				SetControl32BitValue(control, 1);
				GetDialogItemAsControl(ref, index + 6, &control);
				SetControl32BitValue(control, 0);
				printout_pageformat = A4;
				}
		}
	else if(callBackParms->eventData.itemHit == index + 6) {
		if(printout_pageformat == A4) {
				GetDialogItemAsControl(ref, index + 6, &control);
				SetControl32BitValue(control, 1);
				GetDialogItemAsControl(ref, index + 5, &control);
				SetControl32BitValue(control, 0);
				printout_pageformat = LETTER;
				}
		}
	}
return;
}


int MG_GetOutputFName_Plus(char *fname, int maxl, char *dfault, char *message, int use_only_button)
{
    OSErr               anErr = noErr;
    NavReplyRecord      reply;
    NavDialogOptions    dialogOptions;
    OSType              fileTypeToSave = 'TEXT';
    OSType              creatorType;
    NavEventUPP         eventProc = NewNavEventUPP (MyNavEventProc_Plus);
	int    rsult = FALSE;
	static int *p_use;

    anErr = NavGetDefaultDialogOptions (&dialogOptions);
    if (anErr == noErr)
    {
    CtoP(message, dialogOptions.windowTitle);
    dialogOptions.dialogOptionFlags |= kNavNoTypePopup;
    dialogOptions.dialogOptionFlags &= ~ kNavAllowStationery;
    reply.translationNeeded = false;


            //  One way to get the name for the file to be saved.
			CtoP(dfault, dialogOptions.savedFileName);

            creatorType = kNavGenericSignature;
            p_use = &use_only_button;
            anErr = NavPutFile( NULL, &reply, &dialogOptions, eventProc,
                                fileTypeToSave, creatorType, p_use );
            if (anErr == noErr && reply.validRecord)
            {
                AEKeyword   theKeyword;
                DescType    actualType;
                Size        actualSize;
                FSSpec      documentFSSpec;

                anErr = AEGetNthPtr(&(reply.selection), 1, typeFSS,
                                    &theKeyword, &actualType,
                                    &documentFSSpec, sizeof(documentFSSpec),
                                    &actualSize );
                if (anErr == noErr)
                {
                    if (reply.replacing)
                    { /* compute the pathname of the pre-existing file */
                        MG_FSSpecToPathname (&documentFSSpec, fname, maxl);
                        rsult = TRUE;
                    }
                    else
                    { /* the chosen file is new, compute pathname to its directory
                    and append new filename */
                        char empty[256];
                        FSSpec pfss;
                        empty[0] = 0;
                   /* make an FSSpec from volume + parentID of documentFSSpec
                   */
                        FSMakeFSSpec(documentFSSpec.vRefNum, documentFSSpec.parID,
                        	(StringPtr) empty, &pfss);
                   /* compute pathname for this FSSpec
                   */
                        MG_FSSpecToPathname (&pfss, fname, maxl);
                        /* append : new filename to this path
                        */
#if TARGET_RT_MAC_MACHO
                        strcat(fname, "/");
#else
                        strcat(fname, ":");
#endif
                        PtoC(documentFSSpec.name, fname + strlen(fname) );

                        rsult = TRUE;
                    }

                        // DO NOT call NavCompleteSave() to complete
                         anErr = NavCompleteSave(&reply,
                                                kNavTranslateInPlace); 
                }
                (void) NavDisposeReply(&reply);
            }
        DisposeNavEventUPP(eventProc);
    }
    return rsult;
}


char* fl_file_chooser_plus(const char* message, const char* pat, 
	const char* fname, int use_only_button)
{
	static char pathname[FL_PATH_MAX];
	if( MG_GetOutputFName_Plus(pathname, FL_PATH_MAX, (char *)fname, (char *)message, use_only_button) 
		) return pathname;
	else return NULL;
}


#elif defined(WIN32)

extern char *f_format_names[];
extern char *f_format_exts[];
extern int nbr_formats;

char* fl_file_chooser_save_as(const char* message, const char* pat, 
	const char* fname, void *view)
{
int filterindex;
char *outfname, *p, types_list[500];

p = types_list;
for(int f = 0; f < nbr_formats; f++) {
	sprintf(p, "%s format (*.%s)%c*.%s%c", 
		f_format_names[f], f_format_exts[f], 0, f_format_exts[f], 0);
	p += strlen(p) + 1;
	p += strlen(p) + 1;
	}
*p = 0;

filterindex = ((SEA_VIEW *)view)->format_for_save + 1;
outfname = MG_win32_file_chooser(message, 
			types_list,
			fname, 
			memcmp(pat, "*.", 2) == 0 ? pat+2 : NULL,
			TRUE, &filterindex);
if(outfname != NULL) set_save_format((SEA_VIEW *)view, filterindex - 1);
return outfname;
}


#else

class FCB : public Fl_Browser_ {
  void* item_first() const ;
  void* item_next(void*) const ;
  void* item_prev(void*) const ;
  int item_height(const dirent*, int) const ;
  int item_height(void*) const ;
  int item_width(const dirent*) const ;
  int item_width(void*) const ;
  int item_quick_height(void*) const ;
  int incr_height() const ;
  void item_draw(void*, int, int, int, int) const ;
  int checkdir(const dirent*, char*) const ;
  void draw();
  void clear_prev();
public:
  char listed[FL_PATH_MAX];// current dir & starname
  int dirend;		// points after last / before starname
  int nameend;		// length to trailing '*' or '\0'
  const char* pattern;	// default pattern
  dirent** list;	// the file names
  dirent** last;	// pointer after end of list
  const char* message;	// message if no file names
  char preved[FL_PATH_MAX];// directory listed in prev
  dirent** prev;	// cached list of another directory
  dirent** prev_last;	// end of that list
  int prev_count;
  FCB(int x, int y, int w, int h) : Fl_Browser_(x, y, w, h, 0) {
    type(FL_HOLD_BROWSER);
    listed[0] = 0;
    dirend = nameend = 1;
    pattern = 0;
    list = prev = 0;
    message = 0;
  }
  // ~FCB nyi
  void clear();
  void set(const char*);
  int get(char*);
};

class FCW : public Fl_Window {
public:
  int handle(int);
  Fl_Input input;
  Fl_Button* ok_button;
  Fl_Button* cancel_button;
  Fl_Button* normal_button;
  Fl_Button* cwd_button;
  FCB browser;
  FCW() ;
};

#define labelSize 10
#define calc_width(nom) \
	(fl_font(FL_HELVETICA, labelSize), (int)fl_width(nom) + 20)
	

void change_format(Fl_Widget *obj, void *data)
{
SEA_VIEW *view = (SEA_VIEW *)data;
int reponse = ((Fl_Choice *)obj)->value();
if(view->format_for_save != reponse) set_save_format(view, reponse);
return;
}


char* fl_file_chooser_save_as(const char* message, const char* pat, 
	const char* fname, void *data)
{
static FCW* f = NULL; 
static Fl_Choice *choice;
SEA_VIEW *view = (SEA_VIEW *)data;
  
if (!f) {
  f = new FCW();
  f->ok_button->label(fl_ok);
  f->cancel_button->label(fl_cancel);

  if (pat && !*pat) pat = 0;
  if (fname && *fname) {
    f->input.value(fname);
  } else if (f->browser.pattern != pat && (!pat || !f->browser.pattern ||
					   strcmp(pat,f->browser.pattern))) {
    // if pattern is different, remove name but leave old directory:
    const char* p = f->input.value();
    const char* q = fl_filename_name(p);
    f->input.value(p, q-p);
  }
  f->browser.pattern = pat;
  f->normal_button->label(pat ? pat : "visible files");
  f->browser.set(f->input.value());
  f->input.position(10000, f->browser.dirend);

  f->label(message);
  f->hotspot(f);

  static char c_label[] = "Format:";
  choice = new Fl_Choice(2 + calc_width(c_label), f->ok_button->y(), 
	calc_width("Clustal")+25, f->ok_button->h() );
  choice->labelsize(labelSize);
  choice->label(c_label);
  for(int i=0; i < nbr_formats; i++) choice->add(f_format_names[i], 0, 0, 0, 0);
  choice->callback(change_format, view);
  f->add(choice);
  }

choice->value(view->format_for_save);
int before = view->format_for_save;
f->show();
int ok = 0;
for (;;) {
    Fl::wait();
    Fl_Widget* o = Fl::readqueue();
    if (o == f->ok_button) {ok = 1; break;}
    else if (o == f->cancel_button || o == f) break;
  }
f->hide();
f->browser.clear();

if (!ok) {
  	if(view->format_for_save != before) set_save_format(view, before);
  	return 0;
  	}
const char* r = f->input.value();
const char *p;
for (p=r+f->browser.dirend; *p; p++)
    if (*p=='*' || *p=='?' || *p=='[' || *p=='{') return 0;
static char final[FL_PATH_MAX];
strcpy(final, r);
return final;
}

/*

#define PRINTOUT_LABEL_SIZE 30
static char printout_label_block[PRINTOUT_LABEL_SIZE] = "block size=";
static char printout_label_fontsize[PRINTOUT_LABEL_SIZE] = "fontsize=";
static char printout_label_pageformat[PRINTOUT_LABEL_SIZE] = "paper=";
static char printout_label_vary[PRINTOUT_LABEL_SIZE] = "variable sites only";
static char printout_label_black[PRINTOUT_LABEL_SIZE] = "Black&&White";

int change_bouton_label(Fl_Widget *obj, char *message)
{
char *p, defaut[PRINTOUT_LABEL_SIZE], *b_label;
int old_val, new_val;
p = (char *)obj->label();
p = strchr(p, '=') + 1;
sscanf(p, "%d", &old_val);
sprintf(defaut, "%d", old_val);
p = (char *)fl_input(message, defaut);
if(p == NULL) return old_val;
new_val = -1; sscanf(p, "%d", &new_val);
if( new_val == old_val || new_val <= 0) return old_val;
b_label = (char *)malloc(PRINTOUT_LABEL_SIZE);
strcpy(b_label, obj->label());
p = strchr(b_label, '=') + 1;
sprintf(p, "%d", new_val);
free( (void *)obj->label() );
obj->label(b_label);
return new_val;
}

void change_block(Fl_Widget *obj, void *data)
{
printout_block = change_bouton_label(obj, "Enter desired printout block size");
}

void change_fontsize(Fl_Widget *obj, void *data)
{
printout_fontsize = change_bouton_label(obj, 
	"Enter desired fontsize");
}


void change_pageformat(Fl_Widget *obj, void *data)
{
*(int *)data = ((Fl_Choice *)obj)->value();
}

void change_vary(Fl_Widget *obj, void *data)
{
Fl_Button *b = (Fl_Button *)obj;
int etat = b->value();
printout_vary = etat;
}


void change_black(Fl_Widget *obj, void *data)
{
Fl_Button *b = (Fl_Button *)obj;
int etat = b->value();
printout_black = etat;
}

char* fl_file_chooser_plus(const char* message, const char* pat, 
	const char* fname, int use_only_button)
{
static int first = TRUE;
static Fl_Button *b_vary;
  static FCW* f; if (!f) f = new FCW();
  f->ok_button->label(fl_ok);
  f->cancel_button->label(fl_cancel);

  if (pat && !*pat) pat = 0;
  if (fname && *fname) {
    f->input.value(fname);
  } else if (f->browser.pattern != pat && (!pat || !f->browser.pattern ||
					   strcmp(pat,f->browser.pattern))) {
    // if pattern is different, remove name but leave old directory:
    const char* p = f->input.value();
    const char* q = fl_filename_name(p);
    f->input.value(p, q-p);
  }
  f->browser.pattern = pat;
  f->normal_button->label(pat ? pat : "visible files");
  f->browser.set(f->input.value());
  f->input.position(10000, f->browser.dirend);

  f->label(message);
  f->hotspot(f);

if(first) {
	first = FALSE;
int gauche = 2; char *q; Fl_Button *bouton; char *b_label;
Fl_Choice *choice;

f->size(f->w()+calc_width(printout_label_vary)+10, f->h());

bouton = new Fl_Button(gauche, f->ok_button->y(), 5, f->ok_button->h());
b_label = (char *)malloc(50);
strcpy(b_label, printout_label_block);
q = strchr(b_label,'=');
sprintf(q+1, "%d", printout_block);
bouton->size( calc_width(b_label) , bouton->h() );
bouton->label(b_label);
bouton->callback(change_block, NULL);
bouton->labelsize(labelSize);
f->add(bouton);
gauche += bouton->w() + 2;

bouton = new Fl_Button(gauche, f->ok_button->y(), 5, f->ok_button->h());
b_label = (char *)malloc(50);
strcpy(b_label, printout_label_fontsize);
q = strchr(b_label,'=');
sprintf(q+1, "%d", printout_fontsize);
bouton->size( calc_width(b_label) , bouton->h() );
bouton->label(b_label);
bouton->callback(change_fontsize, NULL);
bouton->labelsize(labelSize);
f->add(bouton);
gauche += bouton->w() + 2;

choice = new Fl_Choice(gauche + calc_width(printout_label_pageformat), f->ok_button->y(), 
	calc_width("LETTER")+15, f->ok_button->h() );
choice->labelsize(labelSize);
choice->label(printout_label_pageformat);
choice->add("A4|LETTER");
choice->value(printout_pageformat);
choice->callback(change_pageformat, &printout_pageformat);
f->add(choice);
gauche = choice->x() + choice->w() + 2;

bouton = new Fl_Check_Button(gauche, f->ok_button->y(), 5, 
		f->ok_button->h());
bouton->size( calc_width(printout_label_black) , bouton->h() );
bouton->label(printout_label_black);
bouton->callback(change_black, NULL);
bouton->labelsize(labelSize);
bouton->value(printout_black);
f->add(bouton);
gauche += bouton->w() + 2;

	b_vary = new Fl_Check_Button(gauche, f->ok_button->y(), 5, 
		f->ok_button->h());
	b_vary->size( calc_width(printout_label_vary) , bouton->h() );
	b_vary->label(printout_label_vary);
	b_vary->callback(change_vary, NULL);
	b_vary->labelsize(labelSize);
	b_vary->value(printout_vary);
	f->add(b_vary);
}

f->show();
if(use_only_button) b_vary->show();
else b_vary->hide();

  int ok = 0;
  for (;;) {
    Fl::wait();
    Fl_Widget* o = Fl::readqueue();
    if (o == f->ok_button) {ok = 1; break;}
    else if (o == f->cancel_button || o == f) break;
  }
  f->hide();
  f->browser.clear();

  if (!ok) return 0;
  const char* r = f->input.value();
  const char *p;
  for (p=r+f->browser.dirend; *p; p++)
    if (*p=='*' || *p=='?' || *p=='[' || *p=='{') return 0;
  return (char*)r;
}
*/
#endif
